#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <X11/Xlib.h>

#define N 18
#define O 300
#define R 200
#define L 200
#define RC 50
#define D 300

char *cmnd;

void usage(void) {
    fprintf(stderr, "Usage: %s [-l light_color] [-d dark_color]"
                    " [-t text_color] [-f font] text\n", cmnd);
    exit(1);
}


typedef struct vertexT {
    int x, y, z;
} vertexT;

typedef struct surfaceT {
    vertexT vertex[3];
    XColor xcolor;
} surfaceT;

void DrawObject(
    Display *disp,
    GC gc,
    Drawable d,
    long *zbuf,
    int width,
    int height,
    surfaceT *s,
    size_t n
){
    /* Draw object consisting of |n| surfaces |s| onto drawable |d| using
     * z-buffer |zbuf|.
     * |zbuf| must be initialized (usually with all pixels set to 
     * LONG_MAX) and of the same size as |d|.
     *
     * All coordinates in |s| are assumed to be already pixel
     * coordinates.
     */
    int x, y;
    int i;

    for (i = 0; i < n; i++) {
        vertexT top, left, right, bottom;
        int t;

        #if !FAKE_X
        XSetForeground(disp, gc, s[i].xcolor.pixel);
        #endif
        top = s[i].vertex[0];
        left = s[i].vertex[1];
        bottom = s[i].vertex[2];

        if (top.y > left.y) {
            right = top;
            top = left;
            left = right;
        }
        if (left.y > bottom.y) {
            right = bottom;
            bottom = left;
            left = right;
        }
        if (top.y > left.y) {
            right = top;
            top = left;
            left = right;
        }
        assert(top.y <= left.y);
        assert(left.y <= bottom.y);
        assert(top.y <= bottom.y);      /* paranoia */

        t = bottom.y - top.y;
        if (t == 0) t = 1;

        right.y = left.y;
        right.x = top.x + (bottom.x - top.x) * (right.y - top.y) / t;
        right.z = top.z + (bottom.z - top.z) * (right.y - top.y) / t;
        if (right.x < left.x) {
            vertexT aux;
            aux = right;
            right = left;
            left = aux;
        }
        for (y = top.y; y < left.y; y++) {
            int x0, x1;
            int z0, z1;
            int t;

            t = left.y - top.y;
            if (t == 0) t = 1;
            x0 = top.x + (left.x - top.x) * (y - top.y) / t;
            x1 = top.x + (right.x - top.x) * (y - top.y) / t;
            z0 = top.z + (left.z - top.z) * (y - top.y) / t;
            z1 = top.z + (right.z - top.z) * (y - top.y) / t;
            for (x = x0; x <= x1; x++) {
                int t = (x1 - x0);
                int z = z0 + (z1 - z0) * (x - x0) / ((t != 0) ? t : 1);
                if (z < zbuf[y * width + x]) {
                    zbuf[y * width + x] = z;
                    XDrawPoint(disp, d, gc, x, y);
                }
            }
        }
        for (y = left.y; y <= bottom.y; y++) {
            int x0, x1;
            int z0, z1;
            int t;
            long *zp;

            t = (left.y - bottom.y);
            if (t == 0) t = 1;

            x0 = bottom.x + (left.x - bottom.x) * (y - bottom.y) / t;
            x1 = bottom.x + (right.x - bottom.x) * (y - bottom.y) / t;
            z0 = bottom.z + (left.z - bottom.z) * (y - bottom.y) / t;
            z1 = bottom.z + (right.z - bottom.z) * (y - bottom.y) / t;
            for (x = x0, zp = zbuf + y * width + x0;
                 x <= x1;
                 x++, zp++) {
                int t = x1 - x0;
                int z = z0 + (z1 - z0) * (x - x0) / (t != 0 ? t : 1);
                assert(zp == &zbuf[y * width + x]);
                if (z < *zp) {
                    *zp = z;
                    XDrawPoint(disp, d, gc, x, y);
                }
            }
        }
    }
}

void DrawObjectPerspective(
    Display *disp,
    GC gc,
    Drawable d,
    long *zbuf,
    int width,
    int height,
    surfaceT *s,
    size_t n,
    int ox,
    int oy,
    int dist
){
    /* draw Object perspectivically correct (PC :-) ) as seen from
     * a camera at (0, 0, -dist)
     * Each vertex is projected into the z=0 plane, but only 
     * x and y coordinates are actually changed so that z buffering
     * still works.
     */
    surfaceT *s2;
    int i;

    s2 = malloc(n * sizeof(*s2));

    for (i= 0; i < n; i++) {
        int j;
        s2[i] = s[i];
        for (j = 0; j < 3; j++) {
            s2[i].vertex[j].x = ox +
                                (s2[i].vertex[j].x - ox) * dist /
                                (dist + s2[i].vertex[j].z);
            s2[i].vertex[j].y = oy +
                                (s2[i].vertex[j].y - oy) * dist /
                                (dist + s2[i].vertex[j].z);
        }
    }
    DrawObject(disp, gc, d, zbuf, width, height, s2, n);
    free(s2);
}

void getcolor(Display *disp, surfaceT *s) {
    vertexT a, b, n;
    double r;

    a.x = s->vertex[1].x - s->vertex[0].x;
    a.y = s->vertex[1].y - s->vertex[0].y;
    a.z = s->vertex[1].z - s->vertex[0].z;
    b.x = s->vertex[2].x - s->vertex[0].x;
    b.y = s->vertex[2].y - s->vertex[0].y;
    b.z = s->vertex[2].z - s->vertex[0].z;

    n.x = a.y * b.z - a.z * b.y;
    n.y = a.z * b.x - a.x * b.z;
    n.z = a.x * b.y - a.y * b.x;

    r = ((n.x + n.y - n.z) / sqrt(n.x*n.x + n.y*n.y + n.z*n.z)) * 0.25 + 0.5;
    /* printf ("%g\n", r); */

    s->xcolor.red *= r;
    s->xcolor.green *= r;
    s->xcolor.blue *= r;

    XAllocColor(disp, DefaultColormap(disp, 0), &s->xcolor);
}

int main(int argc, char **argv) {
    Display *disp;
    Window root;
    Window wind;
    XEvent xev;
    GC gc;
    int x, y;
    long *zbuf;
    surfaceT s;
    int i;
    double step;
    int done;
    XColor color;
    struct rusage ru0, ru1;
    struct timeval tv0, tv1;
    int width = 2*O, height = 2*O;
    int ox = O, oy = O;

    cmnd = argv[0];

    disp = XOpenDisplay(NULL);
    
    XSynchronize(disp, 0);

    root = DefaultRootWindow(disp);
 
    wind = XCreateSimpleWindow(
                disp, root,
                0, 0, 2*O, 2*O,
                2, BlackPixel(disp, 0),
                WhitePixel(disp, 0));
    XMapWindow(disp, wind);
    XSelectInput(disp, wind, ExposureMask | KeyPressMask | ButtonPressMask);
    gc = XCreateGC(disp, wind, 0, NULL);

    zbuf = malloc(2*O * 2*O * sizeof(*zbuf));
    for (i = 0; i < 2*O * 2*O; i ++) {
        zbuf[i] = LONG_MAX;
    }
    step = atan(1) / (N/2.0);
    color.red = color.green = color.blue = 65535;

    getrusage(RUSAGE_SELF, &ru0);
    gettimeofday(&tv0, NULL);
    /* draw a sphere */
    for (x = -N; x < N; x++) {
        for (y = 0; y < 4*N; y++) {
            double t;
            s.vertex[0].x = R * sin(x * step) + O;
            t = cos(x * step);
            s.vertex[0].y = R * sin(y * step) * t + O;
            s.vertex[0].z = R * cos(y * step) * t + O;

            s.vertex[1].x = R * sin((x+1) * step) + O;
            t = cos((x+1) * step);
            s.vertex[1].y = R * sin(y * step) * t + O;
            s.vertex[1].z = R * cos(y * step) * t + O;

            s.vertex[2].x = R * sin(x * step) + O;
            t = cos(x * step);
            s.vertex[2].y = R * sin((y+1) * step) * t + O;
            s.vertex[2].z = R * cos((y+1) * step) * t + O;

            s.xcolor = color;
            getcolor(disp, &s);
            DrawObjectPerspective(disp, gc, wind, zbuf, width, height,
            &s, 1, ox, oy, D);


            s.vertex[0].x = R * sin((x+1) * step) + O;
            t = cos((x+1) * step);
            s.vertex[0].y = R * sin((y+1) * step) * t + O;
            s.vertex[0].z = R * cos((y+1) * step) * t + O;

            s.vertex[2].x = R * sin((x+1) * step) + O;
            t = cos((x+1) * step);
            s.vertex[2].y = R * sin(y * step) * t + O;
            s.vertex[2].z = R * cos(y * step) * t + O;

            s.vertex[1].x = R * sin(x * step) + O;
            t = cos(x * step);
            s.vertex[1].y = R * sin((y+1) * step) * t + O;
            s.vertex[1].z = R * cos((y+1) * step) * t + O;

            s.xcolor = color;
            getcolor(disp, &s);
            DrawObjectPerspective(disp, gc, wind, zbuf, width, height,
            &s, 1, ox, oy, D);
        }
    }

    /* draw a cylinder */
    color.red = 65535;
    color.green = color.blue = 0;
    for (y = 0; y < 4*N; y++) {
        s.vertex[0].x = O - L 
            + RC * cos(y * step)
            + RC * sin(y * step);
        s.vertex[0].y = O - L 
            - RC * sin(y * step);
        s.vertex[0].z = O - L 
            - RC * cos(y * step)
            + RC * sin(y * step);
        s.vertex[1].x = O + L 
            + RC * cos(y * step)
            + RC * sin(y * step);
        s.vertex[1].y = O + L 
            - RC * sin(y * step);
        s.vertex[1].z = O + L 
            - RC * cos(y * step)
            + RC * sin(y * step);
        s.vertex[2].x = O + L 
            + RC * cos((y+1) * step)
            + RC * sin((y+1) * step);
        s.vertex[2].y = O + L 
            - RC * sin((y+1) * step);
        s.vertex[2].z = O + L 
            - RC * cos((y+1) * step)
            + RC * sin((y+1) * step);

        s.xcolor = color;
        getcolor(disp, &s);
        DrawObjectPerspective(disp, gc, wind, zbuf, width, height, &s,
        1, ox, oy, D);

        s.vertex[0].x = O + L 
            + RC * cos((y+1) * step)
            + RC * sin((y+1) * step);
        s.vertex[0].y = O + L 
            - RC * sin((y+1) * step);
        s.vertex[0].z = O + L 
            - RC * cos((y+1) * step)
            + RC * sin((y+1) * step);
        s.vertex[1].x = O - L 
            + RC * cos((y+1) * step)
            + RC * sin((y+1) * step);
        s.vertex[1].y = O - L 
            - RC * sin((y+1) * step);
        s.vertex[1].z = O - L 
            - RC * cos((y+1) * step)
            + RC * sin((y+1) * step);
        s.vertex[2].x = O - L 
            + RC * cos(y * step)
            + RC * sin(y * step);
        s.vertex[2].y = O - L 
            - RC * sin(y * step);
        s.vertex[2].z = O - L 
            - RC * cos(y * step)
            + RC * sin(y * step);

        s.xcolor = color;
        getcolor(disp, &s);
        DrawObjectPerspective(disp, gc, wind, zbuf, width, height, &s,
        1, ox, oy, D);


    }
    color.red = 0;
    color.green = color.blue = 65535;
    s.vertex[0].x = O - L;
    s.vertex[0].y = O - L;
    s.vertex[0].z = O - L;
    s.vertex[1].x = O - L + RC;
    s.vertex[1].y = O - L;
    s.vertex[1].z = O - L - RC;
    s.vertex[2].x = O - L + RC;
    s.vertex[2].y = O - L - RC;
    s.vertex[2].z = O - L + RC;

    s.xcolor = color;
    getcolor(disp, &s);
    DrawObjectPerspective(disp, gc, wind, zbuf, width, height, &s, 1,
    ox, oy, D);
    getrusage(RUSAGE_SELF, &ru1);
    gettimeofday(&tv1, NULL);
    printf("%g, %g\n",
           (tv1.tv_sec + tv1.tv_usec * 1E-6)
           -(tv0.tv_sec + tv0.tv_usec * 1E-6),
           (ru1.ru_utime.tv_sec + ru1.ru_utime.tv_usec * 1E-6)
           -(ru0.ru_utime.tv_sec + ru0.ru_utime.tv_usec * 1E-6));

    for (done = 0; !done;) {
        XNextEvent(disp, &xev);
        switch (xev.type) {
        case Expose: 
            break;
        case ButtonPress:
        case KeyPress:
            done = 1;
            break;
        }
    }

    return 0;
}
